/***
Serpent, a Python literal expression serializer/deserializer
(a.k.a. Python's ast.literal_eval in .NET)
Copyright: Irmen de Jong (irmen@razorvine.net)
Software license: "MIT software license". See http://opensource.org/licenses/MIT
***/

namespace Razorvine.Serpent
{
	public static class LibraryVersion
	{
		public const string Version = "1.43";
	}
}
